# Copyright 2018-present Lenovo
# Confidential and Proprietary

import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

import sys
import os
import json
import numpy as np
from itertools import groupby

letrain_path = "/opt/letrain"
path_list = [
    letrain_path + '/applications',
    letrain_path + '/applications/frcnn',
    letrain_path + '/applications/maskrcnn',
    letrain_path + '/applications/maskrcnn/samples',
    letrain_path + '/applications/maskrcnn/samples/coco',
    letrain_path + '/inference',
    letrain_path + '/engine',
]
sys.path += path_list
print(sys.path)

# Import Mask RCNN
from pycocotools.coco import COCO  # noqa: F401
import mrcnn.model as modellib
# Import COCO config
from letrain_keras.config import Config
from lib.utils.timer import Timer

import skimage.io
import skimage.color
from PIL import Image
import tensorflow as tf

# Root directory of the project
ROOT_DIR = os.path.abspath("/tmp")
# Directory to save logs and trained model
MODEL_DIR = os.path.join(ROOT_DIR, "logs")

tf.app.flags.DEFINE_integer(
    "batch_size", 64, "Batch size to use during training.")

FLAGS = tf.app.flags.FLAGS


class CocoConfig(Config):
    """ Configuration for training on MS COCO.
    Derives from the base Config class and overrides values specific
    to the COCO dataset.
    """
    # Give the configuration a recognizable name
    NAME = "coco"

    # We use a GPU with 12GB memory, which can fit two images.
    # Adjust down if you use a smaller GPU.
    IMAGES_PER_GPU = 1

    # Uncomment to train on 8 GPUs (default is 1)
    # GPU_COUNT = 8

    def __init__(self, num_classes, GPU_COUNT=1):
        """Set values of computed attributes."""
        # Number of classes (including background)
        # NUM_CLASSES = 1 + 80  # COCO has 80 classes
        self.NUM_CLASSES = num_classes
        super(CocoConfig, self).__init__()


class InferenceConfig(CocoConfig):
    # Set batch size to 1 since we'll be running inference on
    # one image at a time. Batch size = GPU_COUNT * IMAGES_PER_GPU
    GPU_COUNT = 1
    IMAGES_PER_GPU = 1


def find_last(model_dir):
    """ Find the last chkpnt file of the last trained model.
        Returns: The path of the last checkpoint file
    """
    # Get directory names. Each directory corresponds to a model
    dir_name = os.path.join(model_dir, 'task_0')
    # Find the last checkpoint
    checkpoints = next(os.walk(dir_name))[2]
    checkpoints = filter(lambda f: f.startswith("mask_rcnn"), checkpoints)
    checkpoints = sorted(checkpoints)
    if not checkpoints:
        import errno
        raise IOError(
            errno.ENOENT, "Could not find weight files in {}".format(dir_name))
    checkpoint = os.path.join(dir_name, checkpoints[-1])
    return checkpoint


def maskToRLE(mask):
    """ Convert binary mask to RLE.
        return: RLE (dict {'counts': [], 'size': list(mask.shape)})
    """
    rle = {'counts': [], 'size': list(mask.shape)}
    counts = rle.get('counts')
    for i, (value, elements) in enumerate(groupby(mask.ravel())):
        if i == 0 and value is True:
            counts.append(0)
        counts.append(len(list(elements)))
    return rle


def _parse_results(boxes, masks, class_ids, scores, class_names):
    data = list()
    N = boxes.shape[0]
    for i in range(N):
        if not np.any(boxes[i]):
            continue

        y1, x1, y2, x2 = boxes[i]
        score = round(scores[i].item(), 4)
        class_id = class_ids[i].item()
        label = class_names[class_id]
        data.append(
            {
                "probability": score,
                "label": label,
                "left": x1.item(),
                "top": y1.item(),
                "width": x2.item() - x1.item(),
                "height": y2.item() - y1.item(),
                "rle": maskToRLE(masks[:, :, i])
            }
        )

    return data


class SegmentationDeploySer(object):

    # COCO Class names
    # Index of the class in the list is its ID. For example, to get ID of
    # the teddy bear class, use: class_names.index('teddy bear')

    def __init__(self):
        self.tmp_dir = ""
        self.category_file = 'instances_train.json'
        self.class_names = [
            'BG', 'person', 'bicycle', 'car', 'motorcycle', 'airplane',
            'bus', 'train', 'truck', 'boat', 'traffic light',
            'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird',
            'cat', 'dog', 'horse', 'sheep', 'cow', 'elephant', 'bear',
            'zebra', 'giraffe', 'backpack', 'umbrella', 'handbag', 'tie',
            'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball',
            'kite', 'baseball bat', 'baseball glove', 'skateboard',
            'surfboard', 'tennis racket', 'bottle', 'wine glass', 'cup',
            'fork', 'knife', 'spoon', 'bowl', 'banana', 'apple',
            'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog', 'pizza',
            'donut', 'cake', 'chair', 'couch', 'potted plant', 'bed',
            'dining table', 'toilet', 'tv', 'laptop', 'mouse', 'remote',
            'keyboard', 'cell phone', 'microwave', 'oven', 'toaster',
            'sink', 'refrigerator', 'book', 'clock', 'vase', 'scissors',
            'teddy bear', 'hair drier', 'toothbrush'
        ]

    def load_model(self, model_path, parameters):
        # set batch size, the value is fixed to 1
        FLAGS.batch_size = 1

        # get labels from coco datasets
        coco_annotation_file = os.path.join(model_path, self.category_file)
        if os.path.exists(coco_annotation_file) and \
                os.path.isfile(coco_annotation_file):
            self.class_names = ['background']
            dataset = json.load(open(coco_annotation_file, 'r'))
            assert type(dataset) == dict, 'annotation file format {} not ' \
                                          'supported'.format(type(dataset))
            categories = dataset['categories']
            for cg in categories:
                self.class_names.append(cg['name'])

        # Load image from the images folder, need param "input_dir"
        # file_names = next(os.walk(input_dir))[2]
        config = InferenceConfig(len(self.class_names), 1)
        config.display()

        # Create model object in inference mode.
        self.tmp_dir = os.path.join(os.path.split(model_path)[0], "tmp")
        if not os.path.isdir(self.tmp_dir):
            os.makedirs(self.tmp_dir)
        self.model = modellib.MaskRCNN(
            mode="inference", model_dir=self.tmp_dir, config=config
        )

        # Load weights trained on MS-COCO
        coco_model_path = find_last(model_path)
        self.model.load_weights(coco_model_path, by_name=True)

    def inference(self, image_path):

        # Read image (png image has shape [x,x,4] but expect [x,x,3])
        f_name = os.path.basename(image_path)
        prefix = os.path.splitext(f_name)[0]
        postfix = os.path.splitext(f_name)[1]
        f_dir = os.path.dirname(image_path)
        if postfix in [".png"]:
            im = Image.open(image_path)
            rgb_im = im.convert('RGB')
            image_path = os.path.join(f_dir, prefix + '.jpg')
            rgb_im.save(image_path)
        image = skimage.io.imread(image_path)
        if(len(image.shape) != 3 or image.shape[2] != 3):
            msg = 'Error: {} is not a colorful image.'.format(image_path)
            print(msg)
            if postfix in [".png"]:
                os.remove(os.path.join(f_dir, prefix + '.jpg'))
            raise Exception

        # Run detection
        timer = Timer()
        timer.tic()
        results = self.model.detect([image], verbose=1)
        duration = timer.toc(average=False)
        # Visualize results
        r = results[0]
        # Define response json
        """The Example of the predict_results is:
            {
                "filename": "0001.jpg",
                "duration": 100.90,
                "data": [
                    {
                        "label": "cat",
                        "probability": 0.9,
                        "left": 10,
                        "top": 30,
                        "width": 100,
                        "height": 200
                    }
                ]
            }
        """
        predict_results = {
            "filename": image_path,
            "duration": int(round(1000 * duration)),
            "data": _parse_results(
                r['rois'], r['masks'], r['class_ids'],
                r['scores'], self.class_names
            )
        }

        # Remove extra files
        if postfix in [".png"]:
            os.remove(os.path.join(f_dir, prefix + '.jpg'))
        return predict_results


if __name__ == '__main__':

    parameters = {}
    model_path = '/home/hpcadmin/ai_studio/tasks/'\
                 'fa7d6dc6-043c-11e9-b642-7cd30ae2c038/dada_0/train_dir'
    image_path = '/home/hpcadmin/0_liudanyang/Koala.jpg'
    response_dir = '/home/hpcadmin/0_liudanyang/response_dir'

    seg = SegmentationDeploySer()
    seg.load_model(model_path, parameters)
    seg.inference(image_path)
